from pytest import raises

from triad.utils.schema import (
    quote_name,
    safe_replace_out_of_quote,
    safe_split_and_unquote,
    safe_split_out_of_quote,
    split_quoted_string,
    unquote_name,
)


def test_quote_name():
    assert quote_name("") == "``"
    assert quote_name(" ") == "` `"
    assert quote_name("_") == "`_`"
    assert quote_name("`") == "````"
    assert quote_name("`a`") == "```a```"
    assert quote_name("`a`", quote="'") == "'`a`'"
    assert quote_name("大") == "`大`"
    assert quote_name("`a") == "```a`"

    assert quote_name("a") == "a"
    assert quote_name("Abc") == "Abc"


def test_unquote_name():
    raises(ValueError, lambda: unquote_name(""))
    assert unquote_name("``") == ""
    assert unquote_name("` `") == " "
    assert unquote_name("````") == "`"
    assert unquote_name("`ab") == "`ab"
    assert unquote_name("'大'", quote="'") == "大"
    assert unquote_name("a b") == "a b"

    assert unquote_name("ab") == "ab"


def test_split_qupted_string():
    def _assert(orig, expected):
        res = "".join(
            orig[x[1] : x[2]] if not x[0] else "[" + orig[x[1] : x[2]][1:-1] + "]"
            for x in split_quoted_string(orig, quote="'")
        )
        assert res == expected

    _assert("", "")
    _assert(" ", " ")
    _assert("a''b", "a[]b")
    _assert("a''", "a[]")
    _assert("''b", "[]b")
    _assert("''", "[]")
    _assert("''", "[]")
    _assert("' '", "[ ]")
    _assert("''''", "['']")
    _assert("'''' ''", "[''] []")
    _assert(
        "SELECT * FROM 'tb' WHERE 'tb'.'x''''x'=5",
        "SELECT * FROM [tb] WHERE [tb].[x''''x]=5",
    )
    raises(SyntaxError, lambda: _assert("'", ""))
    raises(SyntaxError, lambda: _assert("'''", ""))
    raises(SyntaxError, lambda: _assert("a 'b. ", ""))


def test_replace():
    assert "" == safe_replace_out_of_quote("", " ", "")
    assert "221`11`1" == safe_replace_out_of_quote("111`11`1", "11", "22")
    assert ".'  '" == safe_replace_out_of_quote(". '  '   ", " ", "", quote="'")
    assert ".''" == safe_replace_out_of_quote(". '  '   ", " ", "", quote="`")


def test_split():
    assert safe_split_out_of_quote("", " ") == [""]
    assert safe_split_out_of_quote(" ", " ") == ["", ""]
    assert safe_split_out_of_quote("  ", " ") == ["", "", ""]
    assert safe_split_out_of_quote("  ", " ", max_split=1) == ["", " "]
    assert safe_split_out_of_quote("  ,", ",") == ["  ", ""]

    assert safe_split_out_of_quote("',',", ",", quote="'") == ["','", ""]
    assert safe_split_out_of_quote("',',',,'", ",", quote="'") == ["','", "',,'"]


def test_split_and_unquote():
    assert safe_split_and_unquote(" a , ' , ' , b ", quote="'") == ["a", " , ", "b"]
    assert safe_split_and_unquote(" a , ' '' ' , b ", quote="'") == ["a", " ' ", "b"]
    # some tolerance of unquoted strings
    assert safe_split_and_unquote(" a , c d , b ", quote="'") == ["a", "c d", "b"]
    assert safe_split_and_unquote(
        " a , '', b, ", quote="'", on_unquoted_empty="keep"
    ) == ["a", "", "b", ""]
    assert safe_split_and_unquote(
        " a , '', b, ", quote="'", on_unquoted_empty="ignore"
    ) == ["a", "", "b"]
    with raises(ValueError):
        safe_split_and_unquote(" a , '', b, ", quote="'", on_unquoted_empty="throw")



from typing import Callable, Dict, Optional, Type, get_type_hints, Any
import inspect
from triad.utils.assertion import assert_or_throw


def _get_first_arg_type(func: Callable) -> Any:
    sig = inspect.signature(func)
    annotations = get_type_hints(func)
    for k, w in sig.parameters.items():
        assert_or_throw(k != "self", ValueError(f"class method is not allowed {func}"))
        assert_or_throw(
            w.kind == w.POSITIONAL_OR_KEYWORD,
            ValueError(f"{w} is not a valid parameter in {func}"),
        )
        anno = annotations.get(k, w.annotation)
        assert_or_throw(
            anno != inspect.Parameter.empty,
            ValueError(f"the first argument must be annotated in {func}"),
        )
        return anno
    raise ValueError(f"{func} does not have any input parameter")


class _ClassExtension:
    def __init__(self, class_type: Type):
        self._class_type = class_type
        self._built_in = set(dir(class_type))
        self._ext: Dict[str, Callable] = {}

    def add_method(
        self, func: Callable, name: Optional[str] = None, on_dup: str = "error"
    ) -> None:
        assert_or_throw(
            name not in self._built_in, ValueError(f"{name} is a built in attribute")
        )
        if name is None:
            name = func.__name__
        if name in self._ext:
            if on_dup == "ignore":
                return
            if on_dup == "error":
                raise ValueError(f"{name} is already registered")
        self._ext[name] = func
        setattr(self._class_type, name, func)


class _ClassExtensions:
    def __init__(self):
        self._types: Dict[Type, _ClassExtension] = {}

    def register_type(self, tp: Type) -> None:
        assert_or_throw(
            tp not in self._types, ValueError(f"{tp} is already registered")
        )
        self._types[tp] = _ClassExtension(tp)

    def add_method(
        self,
        class_type: Type,
        func: Callable,
        name: Optional[str] = None,
        on_dup: str = "error",
    ) -> None:
        assert_or_throw(
            class_type in self._types, ValueError(f"{class_type} is not registered")
        )
        self._types[class_type].add_method(func, name=name, on_dup=on_dup)


_CLASS_EXTENSIONS = _ClassExtensions()


def extensible_class(class_type: Type) -> Type:
    """The decorator making classes extensible by external methods

    :param class_type: the class under the decorator
    :return: the ``class_type``

    .. admonition:: Examples

        .. code-block:: python

            @extensible_class
            class A:

                # It's recommended to implement __getattr__ so that
                # PyLint will not complain about the dynamically added methods
                def __getattr__(self, name):
                    raise NotImplementedError

            @extension_method
            def method(obj:A):
                return 1

            assert 1 == A().method()

    .. note::

        If the method name is already in the original class, a ValueError will be
        thrown. You can't modify any built-in attribute.
    """
    _CLASS_EXTENSIONS.register_type(class_type)
    return class_type


def extension_method(
    func: Optional[Callable] = None,
    class_type: Optional[Type] = None,
    name: Optional[str] = None,
    on_dup: str = "error",
) -> Callable:
    """The decorator to add functions as members of the
    correspondent classes.

    :param func: the function under the decorator
    :param class_type: the parent class type, defaults to None
    :param name: the specified class method name, defaults to None. If None
      then ``func.__name__`` will be used as the method name
    :param on_dup: action on name duplication, defaults to ``error``. ``error``
      will throw a ValueError; ``ignore`` will take no action; ``overwrite``
      will use the current method to overwrite.
    :return: the underlying function

    .. admonition:: Examples

        .. code-block:: python

            @extensible_class
            class A:

                # It's recommended to implement __getattr__ so that
                # PyLint will not complain about the dynamically added methods
                def __getattr__(self, name):
                    raise NotImplementedError

            # The simplest way to use this decorator, the first argument of
            # the method must be annotated, and the annotated type is the
            # class type to add this method to.
            @extension_method
            def method1(obj:A):
                return 1

            assert 1 == A().method1()

            # Or you can be explicit of the class type and the name of the
            # method in the class. In this case, you don't have to annotate
            # the first argument.
            @extension_method(class_type=A, name="m3")
            def method2(obj, b):
                return 2 + b

            assert 5 == A().m3(3)

    .. note::

        If the method name is already in the original class, a ValueError will be
        thrown. You can't modify any built-in attribute.
    """
    if func is not None:  # @extension_method
        _CLASS_EXTENSIONS.add_method(
            _get_first_arg_type(func) if class_type is None else class_type,
            func=func,
            name=name,
            on_dup=on_dup,
        )
        return func
    else:  # @extension_method(...)

        def inner(func):
            _CLASS_EXTENSIONS.add_method(
                _get_first_arg_type(func) if class_type is None else class_type,
                func=func,
                name=name,
                on_dup=on_dup,
            )
            return func

        return inner



from tests.utils.dispatcher_examples import dtest
from triad import run_at_def


@dtest.candidate(lambda x: isinstance(x, int))
def dtest1(a: int) -> int:
    return a


@dtest.candidate(lambda x: isinstance(x, str))
def dtest2(a: str) -> int:
    return len(a)


def dtest3(a: dict) -> int:
    return len(a)


def dtest4(a) -> int:
    return 100


@run_at_def
def register():
    dtest.register(dtest3, lambda x: isinstance(x, dict))


def register2():
    dtest.register(dtest4, lambda x: x is None)



from pytest import raises

from triad.utils.string import assert_triad_var_name, validate_triad_var_name


def test_validate_triad_var_name():
    assert not validate_triad_var_name("")
    assert not validate_triad_var_name("`")
    assert not validate_triad_var_name(" ")
    assert not validate_triad_var_name("中国")
    assert not validate_triad_var_name("_中国")
    assert not validate_triad_var_name("مثال")
    assert not validate_triad_var_name("_")
    assert not validate_triad_var_name("__")

    assert validate_triad_var_name("a")
    assert validate_triad_var_name("abc")
    assert validate_triad_var_name("__a")
    assert validate_triad_var_name("_abc")


def test_assert_triad_var_name():
    raises(AssertionError, lambda: assert_triad_var_name(None))
    raises(AssertionError, lambda: assert_triad_var_name(""))
    raises(AssertionError, lambda: assert_triad_var_name(" "))
    raises(AssertionError, lambda: assert_triad_var_name("1"))
    raises(AssertionError, lambda: assert_triad_var_name(123))
    raises(AssertionError, lambda: assert_triad_var_name("_"))
    raises(AssertionError, lambda: assert_triad_var_name("__"))
    raises(AssertionError, lambda: assert_triad_var_name("a "))
    raises(AssertionError, lambda: assert_triad_var_name("a a"))
    raises(AssertionError, lambda: assert_triad_var_name("大大"))
    raises(AssertionError, lambda: assert_triad_var_name("ŋ"))
    assert "a" == assert_triad_var_name("a")
    assert "_1" == assert_triad_var_name("_1")
    assert "_a_1" == assert_triad_var_name("_a_1")



import fs as pfs
from os.path import exists
import os

from pytest import raises
from triad.collections.fs import FileSystem, _FSPath, _modify_path, _is_windows


def test_modify_path():
    assert "c:/" == _modify_path("/c:")
    assert "s3://" == _modify_path("/s3:")
    assert "C:/" == _modify_path("/C:\\")
    assert "C:/a" == _modify_path("/C:\\a")
    assert "C:/" == _modify_path("/C:\\\\")
    assert "C:/a/b" == _modify_path("/C:\\\\a\\b")
    assert "C:/" == _modify_path("/C:/")
    assert "C:/a" == _modify_path("/C:/a")
    assert "C:/" == _modify_path("/C://")
    assert "C:/a/b" == _modify_path("/C://a/b")

    assert "/" == _modify_path("file://")
    assert "/a/b" == _modify_path("file://a/b")
    assert "c:/a/b" == _modify_path("file:///c:/a/b")
    assert "C:/" == _modify_path("C://")
    assert "c:/x" == _modify_path("c://x")
    assert "c:/" == _modify_path("c:/")
    assert "c:/x" == _modify_path("c:/x")
    assert "c:/" == _modify_path("c:")
    assert "c:/" == _modify_path("c:\\")
    assert "c:/x/" == _modify_path("c:\\x\\")
    raises(NotImplementedError, lambda: _modify_path("\\\\10.0.0.1\1"))


def test_is_windows():
    assert not _is_windows("")
    assert not _is_windows("c")
    assert not _is_windows("c:")
    assert not _is_windows("c:\\")
    assert _is_windows("c:/")
    assert _is_windows("c://")
    assert _is_windows("c:/x")


def test__FSPath():
    p = _FSPath("/a/b.txt")
    assert "" == p.scheme
    assert "/" == p.root
    assert "a/b.txt" == p.relative_path

    p = _FSPath("//a.txt")
    assert "" == p.scheme
    assert "/" == p.root
    assert "a.txt" == p.relative_path

    p = _FSPath("file://a/b.txt")
    assert "" == p.scheme
    assert "/" == p.root
    assert "a/b.txt" == p.relative_path

    p = _FSPath("hdfs://a/b.txt")
    assert "hdfs" == p.scheme
    assert "hdfs://a" == p.root
    assert "b.txt" == p.relative_path

    p = _FSPath("temp://a/b")
    assert "temp" == p.scheme
    assert "temp://a" == p.root
    assert "b" == p.relative_path

    p = _FSPath("/temp:/a/b")
    assert "temp" == p.scheme
    assert "temp://a" == p.root
    assert "b" == p.relative_path
    assert not p.is_windows

    # Windows test cases
    p = _FSPath("c:\\folder\\myfile.txt")
    assert "" == p.scheme
    assert "c:/" == p.root
    assert "folder/myfile.txt" == p.relative_path
    assert p.is_windows

    p = _FSPath("c://folder/myfile.txt")
    assert "" == p.scheme
    assert "c:/" == p.root
    assert "folder/myfile.txt" == p.relative_path
    assert p.is_windows

    p = _FSPath("c:/folder/myfile.txt")
    assert "" == p.scheme
    assert "c:/" == p.root
    assert "folder/myfile.txt" == p.relative_path
    assert p.is_windows

    raises(
        NotImplementedError,
        lambda: _FSPath("\\\\123.123.123.123\\share\\folder\\myfile.txt"),
    )

    raises(ValueError, lambda: _FSPath(None))
    raises(ValueError, lambda: _FSPath(""))
    raises(ValueError, lambda: _FSPath("a.txt"))
    raises(ValueError, lambda: _FSPath("hdfs://"))


def test_fs(tmpdir):
    tmpdir = str(tmpdir)
    # Tests to read and write with tmpdir without FS
    tmpfile = os.path.join(tmpdir, "f.txt")
    f = open(tmpfile, "a")
    f.write("read test")
    f.close()
    f = open(tmpfile, "r")
    assert f.read() == "read test"

    p1 = os.path.join(tmpdir, "a")
    p2 = os.path.join(tmpdir, "b")
    assert not exists(p1)
    assert not exists(p2)
    fs = MockFS()
    fs.makedirs(p1)
    fs.makedirs(p2)
    assert fs.exists(p1) and exists(p1) and os.path.isdir(p1)
    assert fs.exists(p2) and exists(p2) and os.path.isdir(p2)
    assert 1 == fs.create_called
    fs.create_called = 0
    fs.makedirs("temp://x/y")
    fs.makedirs("temp://y/z")
    assert 2 == fs.create_called
    fs.makedirs("mem://x/y")
    fs.makedirs("mem://y/z")
    assert 4 == fs.create_called
    fs.writetext(os.path.join(p1, "a.txt"), "xyz")
    fs.copy(os.path.join(p1, "a.txt"), "mem://y/z/a.txt")
    assert "xyz" == fs.readtext("mem://y/z/a.txt")
    assert not fs.exists("mem://y/z/w/a.txt")
    assert 4 == fs.create_called
    fs.writetext("mem://from/a.txt", "hello")
    fs.copy("mem://from/a.txt", "mem://to/a.txt")
    assert "hello" == fs.readtext("mem://to/a.txt")
    assert 6 == fs.create_called


def test_multiple_writes(tmpdir):
    fs = FileSystem()
    path = os.path.join(tmpdir, "a.txt")
    fs.writetext(path, "1")
    fs.writetext(path, "2")
    assert "2" == fs.readtext(path)

    # auto close is important
    d2 = os.path.join(tmpdir, "x", "y")
    ff = FileSystem(auto_close=False).makedirs(d2, recreate=True)
    ff.writetext("a.txt", "3")
    ff.writetext("a.txt", "4")
    ff = FileSystem(auto_close=False).makedirs(d2, recreate=True)
    ff.writetext("a.txt", "5")
    assert "5" == ff.readtext("a.txt")


def test_glob(tmpdir):
    tmpdir = str(tmpdir)
    fs = FileSystem()
    os.makedirs(os.path.join(str(tmpdir), "d1"))
    os.makedirs(os.path.join(str(tmpdir), "d2", "d2"))
    f = open(os.path.join(str(tmpdir), "d1", "f1.txt"), "a")
    f.write("read test")
    f.close()
    f = open(os.path.join(str(tmpdir), "d2", "d2", "f2.txt"), "a")
    f.write("read test")
    f.close()
    assert {
        pfs.path.join(str(tmpdir), "d1", "f1.txt").replace("\\", "/"),
        pfs.path.join(str(tmpdir), "d2", "d2", "f2.txt").replace("\\", "/"),
    } == {x.path for x in fs.glob("**/*.txt", path=str(tmpdir))}

    fs.makedirs("mem://a/d1")
    fs.makedirs("mem://a/d2/d2")
    fs.touch("mem://a/d1/f3.txt")
    fs.touch("mem://a/d2/d2/f4.txt")
    assert {"mem://a/d1/f3.txt", "mem://a/d2/d2/f4.txt"} == {
        x.path for x in fs.glob("**/*.txt", path="mem://a")
    }


class MockFS(FileSystem):
    def __init__(self):
        super().__init__()
        self.create_called = 0

    def create_fs(self, root):
        self.create_called += 1
        return super().create_fs(root)



from pytest import raises

from triad.utils.rename import _normalize_name, normalize_names


def test_normalize_names():
    raises(ValueError, lambda: normalize_names([]))
    raises(ValueError, lambda: normalize_names([0, 0]))
    raises(ValueError, lambda: normalize_names(["a", 0, "a", 1]))
    raises(ValueError, lambda: normalize_names(["大", "大"]))

    def _assert(input, output):
        res = normalize_names(input)
        assert res == output

    _assert(["a", "b", "c"], {})
    _assert(["a", "+b", "c"], {"+b": "_b"})
    _assert(["a", "+b", "_b"], {"+b": "_b_1"})
    _assert(["a", "+b", "-b", "_b"], {"+b": "_b_1", "-b": "_b_2"})

    _assert([None], {None: "_1"})
    _assert([None, ""], {None: "_1", "": "_2"})
    _assert([None, "_1"], {None: "_1_1"})
    _assert([None, "_1_1", "_1", ""], {None: "_1_1_1", "": "_1_1_2"})

    _assert([0, 2, 1], {0: "_0", 2: "_2", 1: "_1"})

    _assert(["大", "大大"], {"大": "_1", "大大": "_2"})
    _assert(["大a", "大b"], {"大a": "_a", "大b": "_b"})


def test_normalize_name():
    assert _normalize_name("a") == "a"

    assert _normalize_name(None) == ""
    assert _normalize_name("") == ""
    assert _normalize_name("  ") == ""
    assert _normalize_name(" a ") == "a"

    assert _normalize_name("1") == "_1"
    assert _normalize_name("1a") == "_1a"
    assert _normalize_name("1-a") == "_1_a"

    assert _normalize_name("_") == ""

    assert _normalize_name("$%^&") == ""  # no valid chars
    assert _normalize_name("$%(a^b&") == "___a_b_"
    assert _normalize_name("大大a大") == "__a_"
